# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2009 Tiny SPRL (<http://tiny.be>). All Rights Reserved
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

"""stock_move inherit, manage stock moves for check traceability and mixes"""

from osv import osv, fields
from mx.DateTime import now
from datetime import datetime
import netsvc
from tools.translate import _

def reverse_list( item_x, item_y ):
    """ordered a list of ids descently"""
    if item_x > item_y:
        rst = -1
    elif item_x < item_y:
        rst = 1
    else :
        rst = 0
    return rst

class stock_move(osv.osv):
    """stock_move inherit, manage stock moves for check traceability and mixes"""
    _inherit = 'stock.move'

    def write_relationship(self, cr, move_parent_id, move_child_id):
        """method to write to stock_move_history_ids"""
        if move_parent_id and move_child_id:
            if not isinstance(move_parent_id, list):
                move_parent_id = [move_parent_id]
            for parent_id in move_parent_id:
                if parent_id != move_child_id:
                    cr.execute("select parent_id from stock_move_history_ids where parent_id = %s and child_id = %s", (parent_id, move_child_id))
                    if not cr.rowcount:
                        cr.execute('insert into stock_move_history_ids (parent_id, child_id) values (%s,%s)', (parent_id, move_child_id))

        return True

    def get_move_type(self, cr, uid, ids, context=None):
        """return the type of moves"""
        if context is None: context = {}
        res = {}

        for move in self.browse(cr, uid, ids):
            if move.picking_id and move.picking_id.type != 'internal':
                if move.picking_id.type == 'in':
                    res[move.id] = _('IN')
                elif move.picking_id.type == 'out':
                    res[move.id] = _('OUT')
                elif move.picking_id.type == 'return':
                    res[move.id] = _('RETURN')
            else:
                if move.location_dest_id.usage == 'customer':
                    res[move.id] = _('OUT')
                elif move.virtual and move.location_dest_id.usage == 'intermediate':
                    res[move.id] = _('TO MIX')
                elif move.prodlot_id and move.location_id.usage == 'intermediate' and move.prodlot_id.is_mix:
                    res[move.id] = _('MIX')
                elif move.location_dest_id.usage == 'elimination':
                    res[move.id] = _('WASTE')
                elif move.location_dest_id.usage == 'inventory':
                    res[move.id] = _('INVENTORY')
                elif move.location_id.usage == 'inventory':
                    res[move.id] = _('FROM INVENTORY')
                elif move.location_dest_id.usage == 'production':
                    res[move.id] = _('PRODUCTION')
                elif move.production_id:
                    res[move.id] = _('PRODUCTION FINAL MOVE')
                elif move.location_dest_id.usage == 'procurement':
                    res[move.id] = _('PROCUREMENT')
                elif move.procurement_id:
                    res[move.id] = _('PROCUREMENT')
                else:
                    res[move.id] = _('INTERNAL')
        return res


    def _get_supplier(self, cr, uid, ids, field_name, arg, context):
        """gets the name of the supplier"""
        res = {}
        for move in self.browse(cr, uid, ids):
            res[move.id] = None
            if move.location_id.name == 'Suppliers':
                if move.picking_id:
                    res[move.id] = move.picking_id.address_id.partner_id.name
        return res
    
    def _get_expiry_date(self, cr, uid, ids, field_name, arg, context):
        """gets the expiry date of the product"""
        res = {}
        for move in self.browse(cr, uid, ids):
            if move.prodlot_id:
                res[move.id] = move.prodlot_id.dlc
            else:
                res[move.id] = False
        return res

    def _get_moves_to_update_from_prodlot(self, cr, uid, ids, context={}):
        """return moves to update it raises from production lots when update"""
        result = []
        for prodlot_obj_id in self.browse(cr, uid, ids):
            move_ids = self.pool.get('stock.move').search(cr, uid,  [('prodlot_id', '=', prodlot_obj_id.id)])
            result.extend(move_ids)
        return result

    _columns = {
            'supplier': fields.function(_get_supplier, method=True, string="Supplier", type='char', size=64, store = True),
            'expiry_date': fields.function(_get_expiry_date, method=True, string="Expiry Date", type='datetime',
                store = {'stock_production_lot': (_get_moves_to_update_from_prodlot, ['dlc'], 10),
                        'stock.move': (lambda self, cr, uid, ids, c={}: ids, None, 20)}),
            'virtual' : fields.boolean('Virtual Move', help="A virtual move, is an internal move necessary for follow-up the traceability"),
            'procurement_id': fields.many2one('mrp.procurement', 'Procurement', select=True),
            #now this field doesn't use
            'procurements': fields.one2many('mrp.procurement', 'dummy', 'Procurements'),
            'prodlot_id': fields.many2one('stock.production.lot', 'Production Lot', help="Production lot is used to put a serial number on the production", required=False,
            states={'assigned':[('required', True)], 'done':[('required', True)]}),
        }

    def _check_location(self, cr, uid, ids):
        """checks if the location only acepts mixes and if the product is miscible"""
        for move in self.browse(cr, uid, ids):
            if move.location_dest_id.liquid_location and (not move.product_id.liquid) and move.state in ('confirm', 'assigned', 'done'):
                return False
            elif (not move.location_dest_id.liquid_location) and  move.product_id.liquid and 'internal' in move.location_dest_id.usage and move.location_dest_id.chained_location_type == 'none' and move.state in ('confirm', 'assigned', 'done'):
                return False
        return True

    def _check_prodlot(self, cr, uid, ids):
        """checks if prodlot is ok for done move"""
        for move in self.browse(cr, uid, ids):
            if move.state == 'done' and not move.prodlot_id and move.location_id.usage != 'inventory' and move.location_dest_id.usage != 'inventory' and move.product_id.type == 'product':
                return False

        return True

    _constraints = [
    (_check_location, 'Location not valid for this product', ['location_dest_id']),
     (_check_prodlot, 'Production Lot is required to done move', ['prodlot_id'])
    ]
    
    def create_move_mix(self, cr, uid, vals, data):
        """creates the move for the mix"""
        #creates new production lot number
        production_lot_number = self.pool.get('stock.production.lot').make_sscc(cr, uid, context={'product_id': data['obj_virtual_move_inlocation'].product_id.id})
        #creates an object with this production lot number
        mix_vals = {
        'name' : production_lot_number,
        'product_id' : data['obj_virtual_move_tomix'].product_id.id,
        'is_mix' : 1,
        }

        #if prodlots uom is same
        if data['obj_virtual_move_tomix'].prodlot_id.product_uom.id == data['obj_virtual_move_inlocation'].prodlot_id.product_uom.id:
            mix_vals['product_uom'] = data['obj_virtual_move_tomix'].prodlot_id.product_uom.id
            # pylint: disable-msg=W0212
            #if move to mix uom not same that prodlots uom
            if data['obj_virtual_move_tomix'].product_uom.id != mix_vals['product_uom']:
                data['obj_virtual_move_tomix'].product_qty = self.pool.get('product.uom')._compute_qty(cr, uid, data['obj_virtual_move_tomix'].product_uom.id, data['obj_virtual_move_tomix'].product_qty, mix_vals['product_uom'])
            # pylint: disable-msg=W0212
            #if move in location not same that prodlots uom
            if data['obj_virtual_move_inlocation'].product_uom.id != mix_vals['product_uom']:
                data['obj_virtual_move_inlocation'].product_qty = self.pool.get('product.uom')._compute_qty(cr, uid, data['obj_virtual_move_inlocation'].product_uom.id, data['obj_virtual_move_inlocation'].product_qty, mix_vals['product_uom'])

            qty = data['obj_virtual_move_tomix'].product_qty + data['obj_virtual_move_inlocation'].product_qty
            mix_vals['unit_price'] = float(((float(data['obj_virtual_move_tomix'].prodlot_id.unit_price) * float(data['obj_virtual_move_tomix'].product_qty)) + (float(data['obj_virtual_move_inlocation'].prodlot_id.unit_price) * float(data['obj_virtual_move_inlocation'].product_qty))) / float(qty))
        else:
            #default uom from product
            mix_vals['product_uom'] = data['obj_virtual_move_inlocation'].product_id.product_tmpl_id.uom_id.id
            
            
            #TO MIX
            # pylint: disable-msg=W0212
            if data['obj_virtual_move_tomix'].product_uom.id == mix_vals['product_uom']:
                tomix_qty = float(data['obj_virtual_move_tomix'].product_qty)
            else:
                tomix_qty = self.pool.get('product.uom')._compute_qty(cr, uid, data['obj_virtual_move_tomix'].product_uom.id, data['obj_virtual_move_tomix'].product_qty, mix_vals['product_uom'])

            # pylint: disable-msg=W0212
            if data['obj_virtual_move_tomix'].prodlot_id.product_uom.id == mix_vals['product_uom']:
                tomix_price = data['obj_virtual_move_tomix'].prodlot_id.unit_price
            else:
                if data['obj_virtual_move_tomix'].prodlot_id.product_uom.factor_inv_data:
                    tomix_price = data['obj_virtual_move_tomix'].prodlot_id.unit_price / data['obj_virtual_move_tomix'].prodlot_id.product_uom.factor_inv_data
                else:
                    tomix_price = data['obj_virtual_move_tomix'].prodlot_id.unit_price * data['obj_virtual_move_tomix'].prodlot_id.product_uom.factor


            #IN LOCATION
            # pylint: disable-msg=W0212
            if data['obj_virtual_move_inlocation'].product_uom.id == mix_vals['product_uom']:
                inlocation_qty = float(data['obj_virtual_move_inlocation'].product_qty)
            else:
                inlocation_qty = self.pool.get('product.uom')._compute_qty(cr, uid, data['obj_virtual_move_inlocation'].product_uom.id, data['obj_virtual_move_inlocation'].product_qty, mix_vals['product_uom'])

            # pylint: disable-msg=W0212
            if data['obj_virtual_move_inlocation'].prodlot_id.product_uom.id == mix_vals['product_uom']:
                inlocation_price = data['obj_virtual_move_inlocation'].prodlot_id.unit_price
            else:
                if data['obj_virtual_move_inlocation'].prodlot_id.product_uom.factor_inv_data:
                    tomix_price = data['obj_virtual_move_inlocation'].prodlot_id.unit_price / data['obj_virtual_move_inlocation'].prodlot_id.product_uom.factor_inv_data
                else:
                    tomix_price = data['obj_virtual_move_inlocation'].prodlot_id.unit_price * data['obj_virtual_move_inlocation'].prodlot_id.product_uom.factor

            qty = tomix_qty + inlocation_qty

            mix_vals['unit_price'] = float(((float(tomix_price) * float(tomix_qty)) + (float(inlocation_price) * float(inlocation_qty))) / float(qty))

        #create mix lot
        production_lot_number_id = self.pool.get('stock.production.lot').create(cr, uid, mix_vals)

        values = {
          'product_uom' : mix_vals['product_uom'],
          'date' : data['obj_virtual_move_tomix'].date_planned,
          'product_qty' : qty,
          'location_id' : data['obj_virtual_move_tomix'].product_id.product_tmpl_id.property_mix.id,
          'product_id' : data['obj_virtual_move_inlocation'].product_id.id,
          'name' : 'M:' + production_lot_number,
          'date_planned' : data['obj_virtual_move_tomix'].date_planned,
          'state' : 'done',
          'location_dest_id' : data['destination'],
          'prodlot_id' : production_lot_number_id,
        }
        #creates the mix move
        new_id = super(stock_move, self).create(cr, uid, values)
        return new_id
    
    def create_virtual_moves(self, cr, user, obj_move_id, obj_stock_real_prodlots_id, orig_vals):
        """creates the virtual move for the product to mix because need this move that stock decrease"""
        #creates the virtual move from in location move
        date_planned = orig_vals.get('date_planned', False) or obj_move_id.date_planned
        values = {
          'product_uom' : obj_stock_real_prodlots_id.prodlot_id.product_uom.id,
          'product_uos_qty' : obj_stock_real_prodlots_id.name * float(obj_stock_real_prodlots_id.prodlot_id.product_uom.factor),
          'date' : date_planned,
          'product_qty' : obj_stock_real_prodlots_id.name * float(obj_stock_real_prodlots_id.prodlot_id.product_uom.factor),
          'product_uos' : obj_stock_real_prodlots_id.product_id.uos_id and obj_stock_real_prodlots_id.product_id.uos.id or False,
          'location_id' : obj_stock_real_prodlots_id.location_id.id,
          'product_id' : obj_stock_real_prodlots_id.product_id.id,
          'name' : 'M:' + str(obj_stock_real_prodlots_id.location_id.id) + 'TOMVS' + date_planned,
          'date_planned' : date_planned,
          'state' : 'done',
          'location_dest_id' : obj_stock_real_prodlots_id.product_id.product_tmpl_id.property_mix.id,
          'prodlot_id' : obj_stock_real_prodlots_id.prodlot_id.id,
          'virtual' : 1,
        }
        virtual_move_from_inlocation_id = super(stock_move, self).create(cr, user, values)

        #saves the real destination
        destination = orig_vals.get('location_dest_id', False) or obj_move_id.location_dest_id.id
        #changes the destination to property moves for specific product because it's id of virtual location where do mixes
        #vals['location_dest_id'] = move_inlocation.product_id.product_tmpl_id.property_moves.id
        #creates the virtual move from the move to mix

        #movimiento que entra

        vals = {
            'product_uom' : orig_vals.get('location_dest_id', False) or obj_move_id.product_uom.id,
            'product_uos_qty' : orig_vals.get('product_uos_qty', False) or obj_move_id.product_uos_qty,
            'date' : date_planned,
            'product_qty' : orig_vals.get('product_qty', False) or obj_move_id.product_qty,
            'product_uos' : orig_vals.get('product_uos', False) or (obj_move_id.product_uos and obj_move_id.product_uos.id or False),
            'location_id' : orig_vals.get('location_dest_id', False) or obj_move_id.location_dest_id.id,
            'product_id' : orig_vals.get('product_id', False) or obj_move_id.product_id.id,
            'name' : 'M:' + str(obj_move_id.id) + 'TOMVS' + date_planned,
            'date_planned' : date_planned,
            'state' : 'done',
            'location_dest_id' : orig_vals.get('property_mix', False) or obj_move_id.product_id.product_tmpl_id.property_mix.id,
            'prodlot_id' : orig_vals.get('prodlot_id', False) or obj_move_id.prodlot_id.id,
            'virtual' : 1,
            }
        virtual_move_from_movetomix_id = super(stock_move, self).create(cr, user, vals)
        #dictionary with the params to return needed for create the mix (the object of the virtual moves and the real destination)
        data = {
                'obj_virtual_move_inlocation' : self.browse(cr, user, virtual_move_from_inlocation_id),
                'obj_virtual_move_tomix' : self.browse(cr, user, virtual_move_from_movetomix_id),
                'destination' : destination,
                }
        
        return data

    def get_move_in_location(self, cr, uid, location_id, prodlot_id, qty = 0):
        """returns the move in location"""
        moves = []
        ids = self.search(cr, uid, [('location_dest_id', '=', location_id), ('prodlot_id', '=', prodlot_id), ('location_id', '!=', location_id)])
        if len(ids) == 1:
            moves = ids
        elif len(ids) == 0 or not prodlot_id:
            return False
        else:
            ids_to_remove = []
            for move in self.browse(cr, uid, ids):
                m_qty = move.product_qty
                for child_move in move.move_history_ids:
                    m_qty -= child_move.product_qty
                if m_qty <= 0:
                    ids_to_remove.append(move.id)
            if ids_to_remove:
                ids = list(set(ids) - set(ids_to_remove))

            if qty:
                #if there are many items with that params, go on with one param more
                move_id = self.search(cr, uid, [('location_dest_id', '=', location_id), ('prodlot_id', '=', prodlot_id), ('location_id', '!=', location_id), ('product_qty', '>=', qty)])
                if len(move_id) == 1:
                    moves = move_id
                elif len(move_id) > 1:
                    moves = [move_id[len(move_id)-1]]
                else:
                    #check if stock in location are in more that one move
                    select_moves = []
                    move_qty = qty
                    ids.sort(reverse_list)
                    for obj_move_id in self.browse(cr, uid, ids):
                        move_qty = move_qty - obj_move_id.product_qty
                        select_moves.append(obj_move_id.id)
                        if move_qty <= 0:
                            break
                    moves = select_moves

        if moves:
            ids_to_remove = []
            for move in self.browse(cr, uid, moves):
                if move.location_dest_id.usage == 'inventory':
                    ids_to_remove.append(move.id)
            ids = list(set(moves) - set(ids_to_remove))
            return ids
        else:
            return moves
               

    def create(self, cr, uid, vals, context = {}):
        """overwrites this method for check location in purchase orders"""
        if vals.get('purchase_line_id') and vals.get('location_dest_id'):
            obj_purchase_line_id = self.pool.get('purchase.order.line').browse(cr, uid, vals['purchase_line_id'])
            if obj_purchase_line_id.location_id.id != vals['location_dest_id']:
                #El location_id de la compra no es igual al de vals
                vals['location_dest_id'] = obj_purchase_line_id.location_id.id
        return super(stock_move, self).create(cr, uid, vals)

    def write(self, cr, uid, ids, vals, context={}):
        """overwrites this method for create mixes if was necessary"""
        res = []
        if type(ids) != type([]):
            ids = [ids]
        if 'state' in vals:
            if vals['state'] == 'done':
                result = 0
                for move_id in  ids:
                    obj_move_id = self.browse(cr, uid, move_id)
                    #checks if can be a valid move to mix, if the location contents a miscible product of the same type
                    is_mix = False
                    if obj_move_id.location_dest_id.liquid_location and obj_move_id.product_id.liquid and (obj_move_id.prodlot_id or vals.get('prodlot_id', False)) and obj_move_id.product_id.type == 'product':
                        #Es una ubicación de líquidos, el producto es líquido y el movimiento tiene lote
                        res = []
                        res = self.pool.get('stock.real.report.prodlots').search(cr, uid, [('location_id', '=', obj_move_id.location_dest_id.id), ('product_id', '=', obj_move_id.product_id.id)])
                        if len(res) > 0:
                            is_mix = True
                            res2 = self.pool.get('stock.real.report.prodlots').search(cr, uid, [('location_id', '=', obj_move_id.location_dest_id.id), ('prodlot_id', '=', vals.get('prodlot_id', False) or obj_move_id.prodlot_id.id)])
                            if res2:
                                obj_real_report_prodlot_id = self.pool.get('stock.real.report.prodlots').browse(cr, uid, res2[0])
                                if obj_move_id.production_id:
                                    for production_move in obj_move_id.production_id.move_lines:
                                        if production_move.product_uom and production_move.prodlot_id and production_move.location_dest_id:
                                            #obtains the base uom of move uom
                                            base_uom_id = self.pool.get('product.uom').search(cr, uid, [('category_id', '=', production_move.product_uom.category_id.id), ('factor', '=', 1.0)])
                                            #converts the quantity to base uom
                                            production_move.product_qty = self.pool.get('product.uom')._compute_qty(cr, uid, production_move.product_uom.id, production_move.product_qty, base_uom_id[0])
                                            #if all stock in dest location  is consumed in production the location is valid because after confirm production the location was empty
                                            if production_move.prodlot_id.id == obj_real_report_prodlot_id.prodlot_id.id and production_move.product_qty == obj_real_report_prodlot_id.name and production_move.location_id.id == obj_move_id.location_dest_id.id:
                                                is_mix = False
                                                break
                                else:
                                    is_mix = False
                            #Es una mezcla
                            #there are some prodlot in this location
                            if is_mix:
                                for record in res:
                                    obj_record_prodlot_id = self.pool.get('stock.real.report.prodlots').browse(cr, uid, record)
                                    #get the move in location
                                    move_parent_id = self.get_move_in_location(cr, uid, obj_record_prodlot_id.location_id.id, obj_record_prodlot_id.prodlot_id.id, obj_move_id.product_qty)
                                    result = super(stock_move, self).write(cr, uid, move_id, vals, context=context)
                                    if move_parent_id:
                                        params = self.create_virtual_moves(cr, uid, obj_move_id, obj_record_prodlot_id, vals)
                                        mix_id = self.create_move_mix(cr, uid, vals, params)

                                        #creates in move_history_ids the parental relationships if not exists
                                        self.write_relationship(cr, params['obj_virtual_move_inlocation'].id, mix_id)

                                        self.write_relationship(cr, params['obj_virtual_move_tomix'].id, mix_id)

                                        self.write_relationship(cr, move_parent_id, params['obj_virtual_move_inlocation'].id)

                                        self.write_relationship(cr, obj_move_id.id, params['obj_virtual_move_tomix'].id)

                                        move_original_parent_id = self.get_move_in_location(cr, uid, obj_move_id.location_id.id, obj_move_id.prodlot_id.id, obj_move_id.product_qty)
                                        self.write_relationship(cr, move_original_parent_id, obj_move_id.id)

                        if len(res) <= 0 or is_mix == False:
                            #isn't mix
                            res = self.pool.get('stock.real.report.prodlots').search(cr, uid, [('location_id', '=', obj_move_id.location_dest_id.id)])
                            if len(res) == 1:
                                valid = False
                                obj_real_report_prodlot_id = self.pool.get('stock.real.report.prodlots').browse(cr, uid, res[0])
                                if obj_move_id.production_id:
                                    for production_move in obj_move_id.production_id.move_lines:
                                        if production_move.product_uom and production_move.prodlot_id and production_move.location_dest_id:
                                            #obtains the base uom of move uom
                                            base_uom_id = self.pool.get('product.uom').search(cr, uid, [('category_id', '=', production_move.product_uom.category_id.id), ('factor', '=', 1.0)])
                                            #converts the quantity to base uom
                                            production_move.product_qty = self.pool.get('product.uom')._compute_qty(cr, uid, production_move.product_uom.id, production_move.product_qty, base_uom_id[0])
                                            #if all stock in dest location  is consumed in production the location is valid because after confirm production the location was empty
                                            if production_move.prodlot_id.id == obj_real_report_prodlot_id.prodlot_id.id and production_move.product_qty == obj_real_report_prodlot_id.name and production_move.location_id.id == obj_move_id.location_dest_id.id:
                                                valid = True
                                                break
                                if obj_real_report_prodlot_id.prodlot_id.id == vals.get('prodlot_id', False) or obj_move_id.prodlot_id.id:
                                    valid=True
                                if not valid:
                                    raise osv.except_osv(_('Error'),
                                        _("Can not mix products that are not of the same type in this mixtures destination. Only can mix ") + self.pool.get('stock.real.report.prodlots').browse(cr, uid, res[0]).product_id.product_tmpl_id.name + ".")
                            elif (len(res) > 1):
                                raise osv.except_osv(_('Error'), _("Invalid mixtures location. The location has more that one lot."))


                            move_parent_id = self.get_move_in_location(cr, uid, obj_move_id.location_id.id, obj_move_id.prodlot_id.id, obj_move_id.product_qty)
                            self.write_relationship(cr, move_parent_id, obj_move_id.id)
                            result = super(stock_move, self).write(cr, uid, move_id, vals, context=context)
                    elif not obj_move_id.location_dest_id.liquid_location and not obj_move_id.product_id.liquid and (obj_move_id.prodlot_id or vals.get('prodlot_id', False)):
                        res = []
                        #No es una ubicación de líquidos y el producto no es líquido y tiene lote
                        #gets a check if exists a move with all this params
                        if vals.get('prodlot_id', False):
                            res = self.pool.get('stock.real.report.prodlots').search(cr, uid, [('location_id', '=', obj_move_id.location_id.id), ('prodlot_id', '=', vals['prodlot_id'])])
                        elif obj_move_id.prodlot_id:
                            res = self.pool.get('stock.real.report.prodlots').search(cr, uid, [('location_id', '=', obj_move_id.location_id.id), ('prodlot_id', '=', obj_move_id.prodlot_id.id)])
                        if len(res) == 1:
                            obj_record_prodlot_id = self.pool.get('stock.real.report.prodlots').browse(cr, uid, res[0])
                            #gets the move with the params
                            move_parent_ids = self.get_move_in_location(cr, uid, obj_record_prodlot_id.location_id.id, obj_record_prodlot_id.prodlot_id.id, obj_move_id.product_qty)
                            if move_parent_ids:
                                if not isinstance(move_parent_ids, list):
                                    move_parent_ids = [move_parent_ids]
                                for move_parent_id in move_parent_ids:
                                    #creates the move
                                    result = super(stock_move, self).write(cr, uid, move_id, vals, context=context)
                                    #insert the parental move history
                                    if obj_move_id.name.startswith('PROD'):
                                        #Empieza por prod
                                        cr.execute('select parent_id from stock_move_history_ids where child_id = %s', (obj_move_id.id,))
                                        if cr.rowcount:
                                            parent2 = cr.fetchall()
                                            cr.execute('select parent_id from stock_move_history_ids where parent_id = %s and child_id = %s', (move_parent_id, parent2[0][0]))
                                            if not cr.rowcount:
                                                self.write_relationship(cr, move_parent_id, obj_move_id.id)
                                        else:
                                            self.write_relationship(cr, move_parent_id, obj_move_id.id)
                                    else:
                                        self.write_relationship(cr, move_parent_id, obj_move_id.id)
                        elif len(res) > 1:
                            raise osv.except_osv(_('Error'), _("Parental relationship error. The location has more that one lot for this product."))
                        else:
                            result = super(stock_move, self).write(cr, uid, move_id, vals, context=context)
                    elif obj_move_id.prodlot_id or vals.get('prodlot_id', False):
                        #En cualquier otro caso con lote
                        move_parent_id = self.get_move_in_location(cr, uid, obj_move_id.location_id.id, obj_move_id.prodlot_id.id, obj_move_id.product_qty)
                        self.write_relationship(cr, move_parent_id, obj_move_id.id)
                        result = super(stock_move, self).write(cr, uid, move_id, vals, context=context)
                    else:
                        #En cualquier otro caso
                        result = super(stock_move, self).write(cr, uid, move_id, vals, context=context)
                if result:
                    return result
        return super(stock_move, self).write(cr, uid, ids, vals, context=context)

    def action_done(self, cr, uid, ids, context=None):
        """overwrites this method for that the procurement move pass his prodlot to move child"""
        for move in self.browse(cr, uid, ids):
            if move.move_dest_id and (move.state != 'done') and move.prodlot_id and not move.move_dest_id.production_id:
                self.write(cr, uid, move.move_dest_id.id, {'prodlot_id': move.prodlot_id.id})
        return super(stock_move, self).action_done(cr, uid, ids, context)

    def check_assign(self, cr, uid, ids, context={}):
        """overwrites this method beacause if stock in product lot not fix teh procurement in one move there is create more moves"""
        done = []
        count = 0
        pickings = {}
        for move in self.browse(cr, uid, ids):
            if move.product_id.type == 'consu':
                if move.state in ('confirmed', 'waiting'):
                    done.append(move.id)
                pickings[move.picking_id.id] = 1
                continue
            if move.state in ('confirmed', 'waiting'):
                # pylint: disable-msg=W0212
                # warning disabled in pylint because I need access to protected method
                res = self.pool.get('stock.location')._product_reserve(cr, uid, [move.location_id.id], move.product_id.id, move.product_qty, {'uom': move.product_uom.id})
                if res:
                    self.write(cr, uid, move.id, {'state':'assigned'})
                    done.append(move.id)
                    pickings[move.picking_id.id] = 1
                    res_pop = res.pop(0)
                    cr.execute('update stock_move set location_id=%s, product_qty=%s where id=%s', (res_pop[1], res_pop[0], move.id))

                    while res:
                        res_pop = res.pop(0)
                        vals = {
                            'product_uos_qty': move.product_uos_qty,
                            'product_uom': move.product_uom.id or None,
                            'price_unit': move.price_unit,
                            'date':  datetime.now().strftime('%Y-%m-%d  %H:%M:%S'),
                            'product_qty': res_pop[0],
                            'product_uos': move.product_uos.id or None,
                            'location_id': res_pop[1],
                            'product_id': move.product_id.id,
                            'name': move.name,
                            'date_planned': move.date_planned,
                            'picking_id': move.picking_id.id or None,
                            'state': move.state,
                            'location_dest_id': move.location_dest_id.id,
                            'sale_line_id': move.sale_line_id.id,
                            'procurement_id': move.procurement_id.id,
                            'move_dest_id': move.move_dest_id.id or None
                            }
                        move_id = self.create(cr, uid, vals)
                        done.append(move_id)

            if not move.prodlot_id:
                raise osv.except_osv(_('Warning!'),
                                _('Production lots required: Please set reservations\' production lots.'))
        if done:
            count += len(done)
            self.write(cr, uid, done, {'state':'assigned'})

        if count:
            for pick_id in pickings:
                wf_service = netsvc.LocalService("workflow")
                wf_service.trg_write(uid, 'stock.picking', pick_id, cr)
        return count

    def onchange_product_id(self, cr, uid, context, prod_id=False, loc_id=False, loc_dest_id=False, prod_qty=0, prodlot_id=False):
        """event overwrite. Checks if the move not has a realationship with a procurement move and load data for this product"""
        #checks if can write in this move
        result = self.onchange_check_can_write(cr, uid, context=context)
        if result:
            return result
        #return a value with the product and the product_uom, if not product return nothing
        parent_result = super(stock_move, self).onchange_product_id(cr, uid, context, prod_id, loc_id, loc_dest_id)
        if not parent_result:
            #No hay product_id
            return parent_result
        #gets the object of the product
        obj_product_id = self.pool.get('product.product').browse(cr, uid, prod_id)
        #if there is id
        if context and context[0]:
            #checks if this product is available in the location
            result = self.onchange_location_id(cr, uid, context = context, product_id = prod_id, location_id = loc_id, product_qty = prod_qty)
            if 'value' in result and not 'warning' in result:
                if not prodlot_id:
                    result['value']['name'] = parent_result['value']['name']
                else:
                    result['value']['name'] = parent_result['value']['name'] + " (" + self.pool.get('stock.production.lot').browse(cr, uid, prodlot_id).name + ")"
                result['value']['product_uom'] = parent_result['value']['product_uom']
                return result
            elif 'warning' in result:
                return result

        result = {
                    'product_uom': obj_product_id.product_tmpl_id.uom_id.id
                  }

        if prodlot_id:
            result['name'] = obj_product_id.product_tmpl_id.name + " (" + self.pool.get('stock.production.lot').browse(cr, uid, prodlot_id).name + ")"
        else:
            result['name'] = obj_product_id.product_tmpl_id.name
        return {'value': result}

    def onchange_lot_id2(self, cr, uid, ids, prodlot_id=False, product_qty=False, loc_id=False, product_uom=False, context={}):
        """this event manages product_uom and then call real onchange_lot_id"""
        res = self.onchange_lot_id(cr, uid, ids, prodlot_id=prodlot_id, product_qty=product_qty, loc_id=loc_id, context=context)
        if 'warning' in res and res['warning']:
            #Hay warning en lot-id
            return res
        else:
            if prodlot_id:
                prodlot_obj_id = self.pool.get('stock.production.lot').browse(cr, uid, prodlot_id)
                if not 'value' in res:
                    res['value'] = {}

                if product_uom and product_qty:
                    if prodlot_obj_id.product_uom.id != product_uom:
                        # pylint: disable-msg=W0212
                        res['value']['price'] =  prodlot_obj_id.unit_price * self.pool.get('product.uom')._compute_qty(cr, uid, product_uom, product_qty, prodlot_obj_id.product_uom.id)
                    else:
                        res['value']['price'] =  prodlot_obj_id.unit_price * product_qty

        return res

    def onchange_lot_id(self, cr, uid, ids, prodlot_id=False, product_qty=False, loc_id=False, context=None):
        """overwrites this evento for change the name by product code + prodlot name"""
        res = super(stock_move, self).onchange_lot_id(cr, uid, ids, prodlot_id = prodlot_id, product_qty = product_qty, loc_id = loc_id, context = context)
        if 'warning' in res and res['warning']:
            #Hay warning en lot-id
            return res
        else:
            if prodlot_id:
                prodlot_obj_id = self.pool.get('stock.production.lot').browse(cr, uid, prodlot_id)
                if not 'value' in res:
                    res['value'] = {}
                res['value']['name'] = prodlot_obj_id.product_id.product_tmpl_id.name + " (" + prodlot_obj_id.name + ")"

            return res
 
    def onchange_check_can_write(self, cr, uid, context = {}, dummy = False):
        """checks if the move not has a realationship with a procurement move"""
        if context and context[0]:
            obj_move = self.pool.get('stock.move').browse(cr, uid, context[0])
            moves_found = self.search(cr, uid, [('move_dest_id', '=', obj_move.id)])
            if (obj_move.picking_id and obj_move.picking_id.state == 'done') or (moves_found and obj_move.state != 'draft'):
                result = {
                        'location_id': obj_move.location_id.id,
                        'location_dest_id': obj_move.location_dest_id.id,
                        'product_id': obj_move.product_id.id,
                        'product_qty': obj_move.product_qty,
                        'product_uom': obj_move.product_uom.id,
                        'name': obj_move.name,
                        'prodlot_id': obj_move.prodlot_id.id,
                        }
                #shows an alert explains because not changes the location
                warning = {
                    'title': _('Alert !'),
                    'message': _('Can not change because a procurement with a reserve that was created and assigned for this move.')
                    }
                return {'value': result,
                        'warning': warning,}
        return  {}

    def onchange_location_id(self, cr, uid, context, product_id = False, location_id = False, dummy = False, product_qty=False):
        """event fires when changes the location, checks the location and return a default production lot for this location"""
        #if not location or product not return anything
        if not location_id or not product_id:
            return {}
        #if the product quantity needed not exist or is <= 0, set the product quantity to one because is 0 or less not return a valid default_prodlot
        if not product_qty or product_qty <= 0:
            product_qty = 1
        #searches the default prodlot for this location and this product
        default_prodlot = self.pool.get('stock.production.lot').get_default_production_lot(cr, uid, location_id, product_id, product_qty, deep = False)
        #context[0] = stock_move.id
        if context and context[0]:
            obj_move = self.pool.get('stock.move').browse(cr, uid, context[0])
            #if exists prodlot and an id is because the move exists yet, and not can change the location because a procurement created it
            moves_found = self.search(cr, uid, [('move_dest_id', '=', obj_move.id)])
            if (obj_move.picking_id and obj_move.picking_id.state == 'done') or moves_found:
                result = {
                        'location_id': obj_move.location_id.id,
                        'prodlot_id': obj_move.prodlot_id.id,
                        'product_id': obj_move.product_id.id,
                        }
                #shows an alert explains because not changes the location
                warning = {
                    'title': _('Alert !'),
                    'message': _('Can not change because a procurement with a reserve that was created and assigned for this move.')
                    }
                return {'value': result,
                        'warning': warning,}
            #if not return anything, raises an alert because no exist a production lot and must be created
            if not default_prodlot:
                result = {
                        'location_id': obj_move.location_id.id,
                        'prodlot_id': obj_move.prodlot_id.id,
                        'product_id': obj_move.product_id.id,
                        }
                warning = {
                    'title': _('Alert !'),
                    'message': _('There are not any production lot in this location, must be created.')
                    }
                return {'value': result,
                        'warning': warning}
            elif self.pool.get('stock.location').browse(cr, uid, location_id).name == 'Suppliers':
                result = {
                        'prodlot_id': None,
                        }
                return {'value': result}
            else:
                result = {
                        'prodlot_id': default_prodlot,
                        }
                return {'value': result}
        return {}

    def unlink(self, cr, uid, ids, context=None):
        """overwrites this method for managed the deletes moves in move_lines"""
        for move in self.browse(cr, uid, ids, context=context):
            #if the move is a production move with a procurement move
            if move.move_dest_id and move.procurement_id:
                #searches moves that have for move_dest_id this move
                move_found = self.search(cr, uid, [('move_dest_id', '=', move.id)])
                #if have found one move and this move has picking id, this is a procurement move
                if len(move_found) == 1 and self.browse(cr, uid, move_found[0]).picking_id:
                    #gets the procurement id from the move
                    procurement_id = self.pool.get('mrp.procurement').search(cr, uid, [('move_id', '=', move_found[0])])
                    #gets the picking id from the move
                    picking_id = self.browse(cr, uid, move_found[0]).picking_id.id
                    #searches the moves in this picking where the move_id not was the id of move
                    picking_moves = self.search(cr, uid, [('picking_id', '=', picking_id), ('id', '!=', move_found[0])])
                    #if only the move took part of this picking, deletes the picking
                    if len(picking_moves) == 0:
                        self.pool.get('stock.picking').write(cr, uid, picking_id, {'state': 'draft'}, context = context)
                        self.pool.get('stock.picking').unlink(cr, uid, [picking_id])
                    #if only and normally there is a procurement for a procurement move, deletes teh procurement
                    if len(procurement_id) == 1:
                        self.pool.get('mrp.procurement').write(cr, uid, procurement_id[0], {'state': 'draft'}, context = context)
                        self.pool.get('mrp.procurement').unlink(cr, uid, [procurement_id[0]])
                    #addes the procurement move id to move to delete
                    ids.append(move_found[0])
            if move.move_history_ids:
                raise osv.except_osv(_('Error'), _('You cannot remove the move %s which has got child moves !') % (move.name,))

        #updates the moves to draft because in other state not can delete
        self.write(cr, uid, ids, {'state': 'draft'}, context = context)
        #deletes teh moves
        return super(stock_move, self).unlink(cr, uid, ids)

stock_move()
        
                
